import os
import yaml
import shutil
from ..utils.config import get_module_filepath
from ..utils.parsers import sanitize_cloudrun_name

def init_project(args):
    """
    Initializes a new sunholo project with a basic configuration file and directory structure.

**Template Files (`templates/project`):**

A `templates/project` directory is within the `sunholo` package with the following template files in it:

* **`config/llm_config.yaml`:** A basic configuration file with placeholders for LLM settings, vector stores, etc.
* **`config/cloud_run_urls.json`:** A template for Cloud Run URLs.
* **`app.py`:** A basic Flask app that can be customized for the project.
* **`.gitignore`:** A gitignore file to exclude unnecessary files from version control.
* **`README.md`:** A README file with instructions for setting up and running the project.

**Usage:**

Users can initialize a new project using the following command:

```bash
sunholo init my_genai_project
```

This will create a new directory named `my_genai_project` with the template files, allowing users to start building their GenAI application.
    """
    from .sun_rich import console

    project_name = sanitize_cloudrun_name(args.project_name)
    current_dir = os.getcwd()  # This captures the current directory where the command is run
    project_dir = os.path.join(current_dir, project_name)

    console.rule(project_name)
    console.print(f"Initializing in directory: {project_dir}")

    # Create project directory
    if os.path.exists(project_dir):
        console.print(f"[bold red]ERROR: Directory {project_dir} already exists. Please choose a different project name.[/bold red]")
        return

    os.makedirs(project_dir)

    # Copy template files
    template_dir = get_module_filepath("templates/project")
    for filename in os.listdir(template_dir):
        src_path = os.path.join(template_dir, filename)
        dest_path = os.path.join(project_dir, filename)
        if os.path.isfile(src_path):
            shutil.copy(src_path, dest_path)
        elif os.path.isdir(src_path):
            shutil.copytree(src_path, dest_path)
    


    # Determine the location of the generated.tfvars file
    terraform_dir = args.terraform_dir or os.getenv('MULTIVAC_TERRAFORM_DIR')
    if terraform_dir is None:
        raise ValueError("Must specify a terraform_dir or use the MULTIVAC_TERRAFORM_DIR environment variable")
    
    tfvars_file = os.path.join(terraform_dir, 'generated.tfvars')

    # Get the service account, either from the CLI argument or default
    service_account = args.service_account or "sa-llmops"  # Default service account

    # Paths to be included in the cloud build (based on the current working directory)
    # We want paths to start from 'application/system_services/{project_name}'
    relative_base = os.path.relpath(current_dir, os.path.join(current_dir, "..", ".."))
    included_path = os.path.join(relative_base, project_name, "**")
    cloud_build_path = os.path.join(relative_base, project_name, "cloudbuild.yaml")

    update_cloudbuild_template(project_dir, project_name, os.path.join(relative_base, project_name))
    write_vac_config(project_dir, project_name)

    # Define the cloud_run configuration for 'discord-server' with the correct project_dir path
    cloud_run_config = {
        project_name: {
            "cpu": "1",
            "memory": "2Gi",
            "max_instance_count": 3,
            "timeout_seconds": 1500,
            "port": 8080,
            "service_account": service_account,
            "invokers": ["allUsers"],
            "cloud_build": {
                "included": [included_path],
                "path": cloud_build_path,
                "substitutions": {},
                "repo_name": "",
                "repo_owner": ""
            }
        }
    }


    # Initialize the TerraformVarsEditor and update the .tfvars file
    try:
        from ..terraform import TerraformVarsEditor
        editor = TerraformVarsEditor(tfvars_file, terraform_dir)
        editor.update_from_dict(cloud_run_config, 'cloud_run_autogenerated')
    except ImportError as e:
        console.print(f"Error initializing TerraformVarsEditor: {e}")

    from rich.panel import Panel

    console.print(
        Panel((
            "Next steps: \n"
            f" - Navigate to [orange]{project_dir}/config[/orange] and customize VAC configuration files.\n"
            f" - Add your own GenAI app logic to [orange]{project_dir}/vac_service.py[/orange]\n"
            f" - Check terraform [orange]{terraform_dir}/generated.tfvars[/orange] for Multivac deployment"
            ), 
            title=f"Project [bold orange]{project_name}[/bold orange] initialized successfully.",
            subtitle=project_dir),
            )
    console.rule()

def update_cloudbuild_template(project_dir: str, service_name: str, build_folder: str):
    """
    Updates the cloudbuild.yaml template file by replacing the `CHANGE_ME` placeholders with actual values.

    Args:
    -----
    project_dir : str
        The directory where the project and cloudbuild.yaml are located.
    service_name : str
        The name of the service to be used in Cloud Run.
    build_folder : str
        The build folder where the Docker build will take place.

    Example:
    -------
    update_cloudbuild_template('/path/to/project', 'my_service', 'src')
    """
    cloudbuild_path = os.path.join(project_dir, "cloudbuild.yaml")

    # Define the substitutions to replace CHANGE_ME
    substitutions = {
        "_SERVICE_NAME": service_name,
        "_BUILD_FOLDER": build_folder,
    }

    # Read the cloudbuild.yaml template
    with open(cloudbuild_path, 'r') as file:
        content = file.read()

    # Replace each placeholder with its corresponding value
    for placeholder, value in substitutions.items():
        content = content.replace(f"{placeholder}: CHANGE_ME", f"{placeholder}: {value}")


    # Write the updated content back to cloudbuild.yaml
    with open(cloudbuild_path, 'w') as file:
        file.write(content)

    print(f"cloudbuild.yaml updated successfully with service name '{service_name}' and build folder '{build_folder}'.")

def write_vac_config(project_dir: str, service_name: str):
    """
    Writes the vac_config.yaml file with the provided service name as the key.
    """
    vac_config_content = {
        'kind': 'vacConfig',
        'apiVersion': 'v1',
        'vac': {
            service_name: {  # Use the service name here
                'llm': 'vertex',
                'model': 'gemini-1.5-pro-preview-0514',
                'agent': 'vertex-genai',
                'display_name': 'Template VAC',
                'memory': [
                    {
                        'llamaindex-native': {
                            'vectorstore': 'llamaindex'
                        }
                    }
                ],
                'gcp_config': {
                    'project_id': 'llamaindex_project',
                    'location': 'europe-west1',
                    'rag_id': '1234544343434'  # Replace with actual RAG ID
                },
                'chunker': {
                    'chunk_size': 1000,
                    'overlap': 200
                }
            }
        }
    }

    config_dir = os.path.join(project_dir, 'config')
    if not os.path.exists(config_dir):
        os.makedirs(config_dir)

    vac_config_path = os.path.join(config_dir, 'vac_config.yaml')

    # Write the YAML configuration to the file
    with open(vac_config_path, 'w') as file:
        yaml.dump(vac_config_content, file, default_flow_style=False)

    print(f"{vac_config_path} written successfully with service name '{service_name}'.")


def setup_init_subparser(subparsers):
    """
    Sets up an argparse subparser for the 'init' command.
    """
    init_parser = subparsers.add_parser('init', help='Initializes a new Multivac project.')
    init_parser.add_argument('project_name', help='The name of the new project.')
    init_parser.add_argument('--terraform-dir', help='The directory where Terraform files will be generated.')
    init_parser.add_argument('--service-account', help='The service account to use for Cloud Run. Defaults to "sa-llmops"')
    init_parser.set_defaults(func=init_project)



from ..custom_logging import log
from ..utils import ConfigManager

# Load the YAML file
def load_prompt_from_yaml(key, prefix="sunholo", load_from_file=False):
    """
    Returns a string you can use with Langfuse PromptTemplate.from_template() 

    Will first try to load from the Langfuse prompt library, if unavailable will look in promptConfig type file.

    Langfuse prompts have {{ two braces }}, Langchain prompts have { one brace }.

    Example:

    ```python
    from sunholo.langfuse.prompts import load_prompt_from_yaml
    from langchain_core.prompts import PromptTemplate
    
    """
    config = ConfigManager(prefix)
    if load_from_file:
        
        return config.promptConfig(key)


    from langfuse import Langfuse

    # Initialize Langfuse client
    langfuse = Langfuse()

    try:
        if prefix is None:
            langfuse_template = key
        else:
            langfuse_template = f"{prefix}-{key}"
            
        langfuse_prompt = langfuse.get_prompt(langfuse_template, cache_ttl_seconds=300)

        return langfuse_prompt.get_langchain_prompt()
    
    except Exception as err:
        log.warning(f"Could not find langfuse template: {langfuse_template} - {str(err)} - attempting to load from promptConfig")

    return config.promptConfig(key)


import uuid
import base64
import json
from datetime import datetime, timezone

from ..custom_logging import log

def create_metadata(vac, metadata):
    now_utc = datetime.now(timezone.utc)
    formatted_time = now_utc.strftime("%Y-%m-%dT%H:%M:%SZ")

    # Default metadata if none provided
    default_metadata = {"vector_name": vac, "source": "sunholo-cli", "eventTime": formatted_time}

    try:
        # Merge default metadata with provided metadata
        if metadata:
            if not isinstance(metadata, dict):
                metadata = json.loads(metadata)
        else:
            metadata = {}    
    except Exception as err:
        log.error(f"[bold red]ERROR: metadata not parsed: {err} for {metadata}")

    # Update metadata with default values if not present
    metadata.update(default_metadata)

    return metadata

def encode_data(vac, content, metadata=None, local_chunks=False):

    metadata = create_metadata(vac, metadata)

    # Encode the content (URL)
    if isinstance(content, str):
        message_data = base64.b64encode(content.encode('utf-8')).decode('utf-8')
    else:
        raise ValueError(f"Unsupported content type: {type(content)}")

    now_utc = datetime.now(timezone.utc)
    formatted_time = now_utc.strftime("%Y-%m-%dT%H:%M:%SZ")

    # Construct the message dictionary
    messageId = str(uuid.uuid4())
    message = {
        "message": {
            "data": message_data,
            "messageId": messageId,
            "publishTime": formatted_time,
            "attributes": {
                "namespace": vac,
                "return_chunks": str(local_chunks).lower()
            },
        }
    }

    # Merge metadata with attributes
    message["message"]["attributes"].update(metadata)

    #console.print()
    #console.print(f"Sending message: {messageId} with metadata:")
    #console.print(f"{message['message']['attributes']}")

    return message


import json
import uuid
import base64
from datetime import datetime, timezone
from argparse import Namespace
from pathlib import Path

from .sun_rich import console
from rich.progress import Progress

from ..invoke import invoke_vac
from .chat_vac import resolve_service_url
from .run_proxy import stop_proxy
from ..chunker.encode_metadata import create_metadata, encode_data

def embed_command(args):
    chunk_args = vars(args).copy()
    embed_args = vars(args).copy()

    console.rule("Sending data for chunking")

    # Check if the data argument is a file path
    if args.is_file:
        file_path = Path(args.data)
        if not file_path.is_file():
            print(f"ERROR: The specified file does not exist: {file_path}")
            return

    if args.chunk_override:
        chunk_args["url_override"] = args.chunk_override
    else:
        chunk_args["vac_name"] = "chunker"
        chunk_args["url_override"] = ""
    chunk_args = Namespace(**chunk_args)
    chunk_url = resolve_service_url(chunk_args, no_config=True)

    with console.status(f"[bold orange]Sending '{args.data}' to chunk via {chunk_url}[/bold orange]", spinner="star"):
        if args.is_file:

            metadata = create_metadata(args.vac_name, args.metadata)  
            if args.local_chunks:
                metadata["return_chunks"] = True

            chunk_res = invoke_vac(f"{chunk_url}/direct_file_to_embed",
                                   data=file_path,
                                   vector_name=args.vac_name,
                                   metadata=metadata,
                                   is_file=True)
        
        else:
            json_data = encode_data(args.vac_name, args.data, args.metadata, args.local_chunks)
            console.print(f"Chunk JSON data: {json_data}")
            chunk_res = invoke_vac(f"{chunk_url}/pubsub_to_store", json_data)
        
        stop_proxy("chunker")
        if args.only_chunk:
            return chunk_res

        if not args.local_chunks:
            console.rule(f"Chunks sent for processing in cloud: {chunk_res}")

            return
    
    if not chunk_res:
        console.rule(f"[bold orange]No chunks were found for processing of {args.data}[/bold orange]")

        return
    
    console.rule("Processing chunks locally")

    if args.embed_override:
        embed_args["url_override"] = args.embed_override
    else:
        embed_args["vac_name"] = "embedder"
        embed_args["url_override"] = ""
    embed_args = Namespace(**embed_args)
    embed_url = resolve_service_url(embed_args, no_config=True)
    
    if not chunk_res:
        console.print(f"[bold red]ERROR: Did not get any chunks from {chunk_url} for {args.data}")
        stop_proxy("embedder")
        return

    chunks = chunk_res.get('chunks')
    if not chunks:
        console.print(f"[bold red]ERROR: No chunks found within json data: {str(chunk_res)} [/bold red]")
        stop_proxy("embedder")
        return
    
    embeds = []
    with Progress() as progress:
        task = progress.add_task(f"Embedding [{len(chunks)}] chunks via {embed_url}", total=len(chunks))
        for chunk in chunks:
            progress.console.print(f"Working on chunk {chunk['metadata']}")

            # do this async?
            content = chunk.get("page_content")
            now_utc = datetime.now(timezone.utc)
            formatted_time = now_utc.strftime("%Y-%m-%dT%H:%M:%SZ")
            chunk["metadata"]["eventTime"]  = formatted_time
            if not content:
                progress.console.print("[bold red]No content chunk found, skipping.[/bold red]")
                progress.advance(task)
                continue
            progress.console.print(f"Sending chunk length {len(content)} to embedder")
            processed_chunk = encode_data(vac = args.vac_name, 
                                          content = json.dumps(chunk))
            
            embed_res = invoke_vac(f"{embed_url}/embed_chunk", processed_chunk)
            embeds.append(embed_res)
            progress.advance(task)

    stop_proxy("embedder")
    console.rule("Embedding pipeline finished")
    
    return embed_res


def setup_embedder_subparser(subparsers):
    """
    Sets up an argparse subparser for the 'embed' command.

    Args:
        subparsers: The subparsers object from argparse.ArgumentParser().
    """
    embed_parser = subparsers.add_parser('embed', help='Send data for embedding to a VAC vector store')
    embed_parser.add_argument('--embed-override', help='Override the embed VAC service URL.')
    embed_parser.add_argument('--chunk-override', help='Override the chunk VAC service URL.')
    embed_parser.add_argument('--no-proxy', action='store_true', help='Do not use the proxy and connect directly to the VAC service.')
    embed_parser.add_argument('-m', '--metadata', default=None, help='Metadata to send with the embedding (as JSON string).')
    embed_parser.add_argument('--local-chunks',  action='store_true', help='Whether to process chunks to embed locally, or via the cloud.')
    embed_parser.add_argument('vac_name', help='VAC service to embed the data for')
    embed_parser.add_argument('data', help='String content to send for embedding')
    embed_parser.add_argument('--is-file', action='store_true', help='Indicate if the data argument is a file path')
    embed_parser.add_argument('--only-chunk', action='store_true', help='Whether to only parse the document and return the chunks locally, with no embedding')

    embed_parser.set_defaults(func=embed_command)



from ..custom_logging import log
from ..agents import send_to_qa
from ..qna.parsers import parse_output
from ..streaming import generate_proxy_stream
from ..utils import ConfigManager
from ..utils.api_key import has_multivac_api_key

def direct_vac(vac_input: dict, vac_name: str, chat_history=[]):
    """
    This lets VACs call other VAC Q&A endpoints within their code
    """

    log.info(f"Invoking VAC Q&A endpoints for {vac_name}")

    if 'user_input' not in vac_input:
        raise ValueError(f'vac_input must contain at least "user_input" key - got {vac_input}')

    global_config = ConfigManager('global')
    config = ConfigManager(vac_name)

    agent_name = config.vacConfig('agent')
    agent_url = config.vacConfig("agent_url")

    if agent_url:
        log.info("Found agent_url within vacConfig: {agent_url}")
    # via public cloud endpoints - assumes no gcloud auth
    override_endpoint = None
    if has_multivac_api_key():
        print("Found MULTIVAC_API_KEY")
        gcp_config = global_config.vacConfig("gcp_config")
        endpoints_base_url = gcp_config.get("endpoints_base_url")
        if not endpoints_base_url:
            raise ValueError("MULTIVAC_API_KEY env var is set but no config.gcp_config.endpoints_base_url can be found")

        override_endpoint = f"{endpoints_base_url}/v1/{agent_name}"

    override_endpoint = agent_url or override_endpoint

    print(f"Using {override_endpoint=}")
    
    # Prepare the kwargs for send_to_qa by copying vac_input and adding more values
    qa_kwargs = vac_input.copy()

    # Add additional arguments
    qa_kwargs.update({
        'vector_name': vac_name,
        'chat_history': chat_history,
        'image_url': vac_input.get('image_url') or vac_input.get('image_uri'),
        'override_endpoint': override_endpoint,
        'message_source': "sunholo.invoke_vac_qa.invoke",
        'stream': False,
        'configurable': {
            "vector_name": vac_name,
        },
    })

    log.info(f'Batch invoke_vac_qa {vac_name} with {qa_kwargs=}')

    vac_response = send_to_qa(**qa_kwargs)
        
    # ensures {'answer': answer}
    answer = parse_output(vac_response)
    chat_history.append({"name": "Human", "content": vac_input})
    chat_history.append({"name": "AI", "content": answer})
    answer["chat_history"] = chat_history
    
    return answer

def direct_vac_stream(vac_input: dict, vac_name: str, chat_history=[]):

    if 'user_input' not in vac_input:
        raise ValueError('vac_input must contain at least "user_input" key - got {vac_input}')

    user_id = vac_input.get('user_id')
    session_id = vac_input.get('session_id')
    image_uri = vac_input.get('image_url') or vac_input.get('image_uri')
    
    log.info(f"Streaming invoke_vac_qa with {vac_input=}")
    def stream_response():
        generate = generate_proxy_stream(
                send_to_qa,
                vac_input["user_input"],
                vector_name=vac_name,
                chat_history=chat_history,
                generate_f_output=lambda x: x,  # Replace with actual processing function
                stream_wait_time=0.5,
                stream_timeout=120,
                message_author=user_id,
                #TODO: populate these
                image_url=image_uri,
                source_filters=None,
                search_kwargs=None,
                private_docs=None,
                whole_document=False,
                source_filters_and_or=False,
                # system kwargs
                configurable={
                    "vector_name": vac_name,
                },
                user_id=user_id,
                session_id=session_id, 
                message_source="sunholo.invoke_vac_qa.stream"
        )
        for part in generate():
            yield part

    answer = ""

    for token in stream_response():
        if isinstance(token, bytes):
            token = token.decode('utf-8')
            yield token
        if isinstance(token, dict):
            # ?
            pass
        elif isinstance(token, str):
            answer += token

    if answer:
        chat_history.append({"name": "Human", "content": vac_input})
        chat_history.append({"name": "AI", "content": answer})

    return chat_history


from __future__ import annotations
from typing import Optional, Tuple, TYPE_CHECKING
if TYPE_CHECKING:
    from PIL import Image
    from google.cloud.storage.bucket import Bucket

import os
from urllib.parse import quote
from datetime import datetime, timedelta

# needs to be in minimal to check gcp
from google.auth.exceptions import RefreshError

try:
    from google.cloud import storage
except ImportError:
    storage = None

from ..custom_logging import log
from ..auth.refresh import refresh_credentials, get_default_email
from io import BytesIO
try:
    from PIL import Image
except ImportError:
    Image = None


def get_image_from_gcs(gs_uri: str) -> Image.Image: # type: ignore
    """Converts image bytes from GCS to a PIL Image object."""
    image_bytes = get_bytes_from_gcs(gs_uri)
    if not Image:
        raise ImportError('Could not import PIL (pillow) - install via `pip install sunholo[gcp]`')
    try:
        img = Image.open(BytesIO(image_bytes))
        return img
    except IOError as e:
        raise ValueError("Unable to open image from bytes:", e)

def get_bytes_from_gcs(gs_uri) -> Optional[bytes]:
    """
    Downloads a file from Google Cloud Storage and returns its bytes.

    Args:
        gs_uri (str): The Google Cloud Storage URI of the file to download (e.g., 'gs://bucket_name/file_name').

    Returns:
        bytes: The content of the file in bytes, or None if an error occurs.
    """
    if not gs_uri.startswith('gs://'):
        log.error(f"Invalid GCS URI: {gs_uri}")
        return None
    
    try:
        storage_client = storage.Client()
    except Exception as err:
        log.error(f"Error creating storage client: {str(err)}")
        return None
    
    try:
        # Parse the GCS URI
        path_parts = gs_uri[5:].split('/', 1)
        bucket_name = path_parts[0]
        blob_name = path_parts[1]

        # Get the bucket and blob
        bucket = storage_client.bucket(bucket_name)
        blob = bucket.blob(blob_name)

        # Download the blob as bytes
        file_bytes = blob.download_as_bytes()
        return file_bytes

    except Exception as err:
        log.error(f"Error downloading file from GCS: {str(err)}")
        return None

gcs_bucket_cache = {}
def get_bucket(bucket_name: str) -> Optional[Bucket]:
    """
    Gets a Cloud Storage bucket and initialised GCS client

    Args:
        bucket_name: Name of the bucket 

    Returns:

    """
    if storage:
        gcs_client = storage.Client()
    else:
        raise ImportError("google storage pip required - install via `pip install sunholo[gcp]`")

    if bucket_name not in gcs_bucket_cache:
        gcs_bucket_cache[bucket_name] = gcs_client.get_bucket(bucket_name)
    return gcs_bucket_cache[bucket_name]

def sign_gcs_url(bucket_name:str, object_name:str, expiry_secs:int = 86400) -> Optional[str]:
    """
    Creates a signed URL so that users can download a file from Google Cloud Storage without authentication

    Args:
        bucket_name: Name of the bucket where the object lies
        object_name: Object within the bucket
        expiry_secs: How long the link will be valid - default 24hrs
    
    Returns:
        str: The signed URL or None if not avialable
    """
    service_account_email, gcs_credentials = get_default_email()

    expires = datetime.now() + timedelta(seconds=expiry_secs)

    if object_name.startswith("gs://"):
        object_name = object_name.replace(f"gs://{bucket_name}/","")

    try:
        bucket = get_bucket(bucket_name)
        blob = bucket.blob(object_name)
        if not blob:
            return None
        url = blob.generate_signed_url(
            version="v4",
            expiration=expires,
            service_account_email=service_account_email,
            access_token=gcs_credentials.token)
        return url
    except RefreshError:
        log.info("Refreshing gcs_credentials due to token expiration.")
        credentials, token = refresh_credentials()
        if credentials:
            return sign_gcs_url(bucket_name, object_name, expiry_secs)
        log.error("Failed to refresh gcs credentials")
        return None
    except Exception as e:
        log.error(f"Failed to generate signed URL: {e}")
        return None


def construct_download_link(source_uri: str) -> Tuple[str, str, bool]:
    """Creates a viewable Cloud Storage web browser link from a gs:// URI.""" 
    if not source_uri.startswith("gs://"):
        return source_uri, source_uri, False  # Return the URI if it doesn't start with gs://
    
    bucket_name, object_name = parse_gs_uri(source_uri)
    log.info(f"Source URL: {source_uri} parsed to {bucket_name=} - {object_name=}")

    signed_url = sign_gcs_url(bucket_name, object_name)
    if signed_url:
        the_name = os.path.basename(object_name)
        log.info(f"Creating signed URL for {the_name} - {signed_url}")
        return signed_url, the_name, True
    
    log.error(f"Failed to generate signed URL for {source_uri}")
    return construct_download_link_simple(bucket_name, object_name)


def construct_download_link_simple(bucket_name:str, object_name:str) -> Tuple[str, str, bool]:
    """Creates a viewable Cloud Storage web browser link from a gs:// URI.

    Args:
        source_uri: The gs:// URI of the object in Cloud Storage.

    Returns:
        A URL that directly access the object in the Cloud Storage web browser.
    """

    if object_name.startswith("gs://"):
        public_url = object_name.replace("gs://", "https://storage.cloud.google.com")
    else:
        public_url = f"https://storage.cloud.google.com/{bucket_name}/{quote(object_name)}"
    filename = os.path.basename(object_name)
    return public_url, filename, False

def parse_gs_uri(gs_uri: str) -> Tuple[str, str]:
    """Parses a gs:// URI into the bucket name and object name.

    Args:
        gs_uri: The gs:// URI to parse.

    Returns:
        A tuple containing the bucket name and object name.
    """
    parts = gs_uri.split("/")
    if len(parts) < 3:
        raise ValueError(f"Invalid gs:// URI: {gs_uri}")
    return parts[2], "/".join(parts[3:])


import traceback

from ..custom_logging import log
from ..utils import ConfigManager
from .safety import genai_safety

from typing import TYPE_CHECKING, Union

import json

try:
    import google.generativeai as genai
except ImportError:
    genai = None

if TYPE_CHECKING:
    from google.generativeai.protos import Part

class GenAIFunctionProcessor:
    """
    A generic class for processing function calls from google.generativeai function calling models.

    This class provides a framework for handling multiple function calls in responses
    from generative AI systems. Users of this class should subclass it and provide
    their own implementation of the `construct_tools` method, which returns a dictionary
    of function names mapped to their implementations.

    Attributes:
        config (ConfigManager): Configuration manager instance. Reach values via self.config within your own construct_tools() method
        funcs (dict): A dictionary of function names mapped to their implementations.

    Example usage:

    ```python
    class AlloyDBFunctionProcessor(GenAIFunctionProcessor):
        def construct_tools(self) -> dict:
            pass

    config = ConfigManager()
    alloydb_processor = AlloyDBFunctionProcessor(config)

    results = alloydb_processor.process_funcs(full_response)

    alloydb_model = alloydb_processor.get_model(
        model_name="gemini-1.5-pro",
        system_instruction="You are a helpful AlloyDB agent that helps users search and extract documents from the database."
    )
    ```
    """

    def __init__(self, config: ConfigManager):
        """
        Initializes the GenAIFunctionProcessor with the given configuration.

        Args:
            config (ConfigManager): The configuration manager instance.
        """
        if not genai:
            raise ImportError("import google.generativeai as genai is required, import via `pip install sunholo[gcp]`")
        
        self.config = config
        self.funcs = self.construct_tools()
        self.model_name = config.vacConfig("model") if config.vacConfig("llm") == "vertex" else "gemini-1.5-flash"
        self.last_api_requests_and_responses = []
        self._validate_functions()

    def construct_tools(self) -> dict:
        """
        Constructs a dictionary of tools (functions) specific to the application.

        This method should be overridden in subclasses to provide the specific
        function implementations required for the application.

        Returns:
            dict: A dictionary where keys are function names and values are function objects

        Raises:
            NotImplementedError: If the method is not overridden in a subclass.
        """
        raise NotImplementedError("Subclasses must implement this method to return a dictionary of functions.")

    def _validate_functions(self):
        """
        Validates that all functions in the `funcs` dictionary have docstrings.

        This method checks each function in the `funcs` dictionary to ensure it has
        a docstring. If a function is missing a docstring, an error is logged, and
        a `ValueError` is raised.

        Raises:
            ValueError: If any function is missing a docstring.
        """
        for func_name, func in self.funcs.items():
            if not func.__doc__:
                log.error(f"Function {func_name} is missing a docstring.")
                raise ValueError(f"Function {func_name} must have a docstring to be used as a genai tool.")
    

    def parse_as_parts(self, api_requests_and_responses=[]):
        if not api_requests_and_responses and not self.last_api_requests_and_responses:
            log.info("No api_requests_and_responses found to parse to parts")
            return None

        if api_requests_and_responses:
            self.last_api_requests_and_responses = api_requests_and_responses
        
        from google.generativeai.protos import Part

        work_on = api_requests_and_responses or self.last_api_requests_and_responses
        parts = []
        for part in work_on:
            try:
                parts.append(
                    Part(
                        function_response=genai.protos.FunctionResponse(
                            name=part[0],
                            response={"result": part[2], "args": json.dumps(part[1])}
                        )
                    )
                )
            except Exception as err:
                parts.append(
                    Part(
                        function_response=genai.protos.FunctionResponse(
                            name=part[0],
                            response={"result": f"ERROR: {str(err)}"}
                        )
                    )
                )                

        return parts

    def check_function_result(self, function_name, target_value, api_requests_and_responses=[]):
        """
        Checks if a specific function result in the api_requests_and_responses contains a certain value.

        Args:
            function_name (str): The name of the function to check.
            target_value: The value to look for in the function result.
            api_requests_and_responses (list, optional): List of function call results to check. 
                                                        If not provided, the method will use `self.last_api_requests_and_responses`.

        Returns:
            bool: True if the target_value is found in the specified function's result, otherwise False.
        """
        if not api_requests_and_responses:
            api_requests_and_responses = self.last_api_requests_and_responses

        if not api_requests_and_responses:
            log.info("No api_requests_and_responses found to check.")
            return False

        for part in api_requests_and_responses:
            func_name = part[0]
            result = part[2]

            if func_name == function_name:
                if isinstance(result, list) and target_value in result:
                    log.info(f"Target value '{target_value}' found in the result of function '{function_name}'.")
                    return True
                elif isinstance(result, dict) and isinstance(target_value, dict):
                    for key, expected_value in target_value.items():
                        if key in result:
                            if result[key] == expected_value:
                                log.info(f"The key '{key}' has the same value in both dictionaries.")
                                return True
                    return False
                elif result == target_value:
                    log.info(f"Target value '{target_value}' found in the result of function '{function_name}'.")
                    return True

        log.info(f"Target value '{target_value}' not found in the result of function '{function_name}'.")
        return False
    
    def parse_as_string(self, api_requests_and_responses=[]):
        if not api_requests_and_responses and not self.last_api_requests_and_responses:
            log.info("No api_requests_and_response found to parse to string")
            return None
        
        if api_requests_and_responses:
            self.last_api_requests_and_responses = api_requests_and_responses
        
        work_on = api_requests_and_responses or self.last_api_requests_and_responses
        strings = []
        for part in work_on:
            strings.append(
                f"function tool {part[0]} was called with arguments: {part[1]} and got this result:\n"
                f"<{part[0]}_result>{part[2]}</{part[0]}_result>"
            )

        return strings

    def process_funcs(self, full_response, output_parts=True) -> Union[list['Part'], str]:
        """
        Processes the functions based on the full_response from the generative model.

        This method iterates through each part of the response, extracts function
        calls and their parameters, and executes the corresponding functions defined
        in the `funcs` dictionary.

        Args:
            full_response: The response object containing function calls.
            output_parts (bool): Indicates whether to return structured parts or plain strings.

        Returns:
            list[Part] | str: A list of Part objects or a formatted string with the results.

        Example usage:
        ```python
        results = alloydb_processor.process_funcs(full_response)
        ```
        """
        api_requests_and_responses = []

        if not full_response:
            log.info("No response was found to process")
            return api_requests_and_responses

        # Loop through each part in the response to handle multiple function calls
        #TODO: async
        for part in full_response.candidates[0].content.parts:
            if fn := part.function_call:
                # Extract parameters for the function call
                function_name = fn.name
                params_obj = {key: val for key, val in fn.args.items()}
                params = ', '.join(f'{key}={val}' for key, val in params_obj.items())
                log.info(f"Executing {function_name} with params {params}")

                # Check if the function is in our dictionary of available functions
                if function_name in self.funcs:
                    try:
                        # Execute the function with the provided parameters
                        result = self.funcs[function_name](**params_obj)
                        log.info(f"Got result from {function_name}: {result}")
                    except Exception as err:
                        error_message = f"Error in {function_name}: {str(err)}"
                        traceback_details = traceback.format_exc()
                        log.warning(f"{error_message}\nTraceback: {traceback_details}")
                        result = [f"{error_message}\n{traceback_details}"]

                    api_requests_and_responses.append(
                        [function_name, params, result]
                    )
                else:
                    log.error(f"Function {function_name} is not recognized")

        log.info(f"{api_requests_and_responses=}")
        self.last_api_requests_and_responses = api_requests_and_responses

        if output_parts:
            return self.parse_as_parts()
        
        return self.parse_as_string()
    
    def tool_config_setting(self, mode:str):
        from google.generativeai.types import content_types

        fns = list(self.funcs.keys())

        if fns and mode == "any":
            return content_types.to_tool_config(
                    {"function_calling_config": {"mode": mode, "allowed_function_names": fns}}
                )
        else:
            return content_types.to_tool_config(
                {"function_calling_config": {"mode": mode}}
            )

    def get_model(
            self, 
            system_instruction: str, 
            generation_config=None, 
            model_name: str=None,
            tool_config: str="auto"):
        """
        Constructs and returns the generative AI model configured with the tools.

        This method creates a generative AI model using the tools defined in the
        `funcs` dictionary and the provided configuration options.

        Args:
            model_name (str): The name of the model to use.
            system_instruction (str): Instructions for the AI system.
            generation_config (dict, optional): Configuration for generation, such as temperature.
            tool_config (str, optional): Configuration for tool behaviour: 'auto' it decides, 'none' no tools, 'any' always use tools

        Returns:
            GenerativeModel: An instance of the GenerativeModel configured with the provided tools.

        Example usage:

        ```python
        alloydb_model = alloydb_processor.get_model(
            model_name="gemini-1.5-pro",
            system_instruction="You are a helpful AlloyDB agent that helps users search and extract documents from the database."
        )
        ```
        """
        if generation_config is None:
            generation_config = {
                "temperature": 0.1,
                "max_output_tokens": 8000,
            }

        # Extract the functions from the dictionary to pass into the model
        tools = list(self.funcs.values())

        try:
            model = genai.GenerativeModel(
                model_name=model_name or self.model_name,
                tools=tools,
                tool_config=self.tool_config_setting(tool_config),
                generation_config=generation_config,
                safety_settings=genai_safety(),
                system_instruction=system_instruction,
            )
            return model
        except Exception as err:
            log.error(f"Error initializing model: {str(err)}")
            return None


import os
from subprocess import Popen
from ..utils.config import load_all_configs

def deploy_vac(args):
    """
    Deploys the VAC by running a Flask app locally.
    """
    print(f"Deploying VAC: {args.vac_name} locally")

    # Load the vacConfig
    configs_by_kind = load_all_configs()
    vac_config = configs_by_kind.get('vacConfig', {}).get('vac', {}).get(args.vac_name)

    if not vac_config:
        raise ValueError(f"No configuration found for VAC: {args.vac_name}")

    # Assuming the Flask app is in 'app.py' within the config path
    app_path = os.path.join(args.config_path, 'app.py')
    if not os.path.exists(app_path):
        raise ValueError(f"app.py not found in {args.config_path}")

    print(f"Running Flask app from {app_path}")

    # Run the Flask app
    command = ["python", app_path]
    print(f"Running Flask app with command: {' '.join(command)}")
    process = Popen(command)
    process.communicate()

def setup_deploy_subparser(subparsers):
    """
    Sets up an argparse subparser for the 'deploy' command.

    Example command:
    ```bash
    sunholo deploy "vac_name" --config_path .
    ```
    """
    deploy_parser = subparsers.add_parser('deploy', help='Triggers a deployment of a VAC.')
    deploy_parser.add_argument('vac_name', help='The name of the VAC to deploy.')
    deploy_parser.add_argument('--config_path', default='.', help='Path to the directory containing the config folder `config/` and Flask app `app.py`, defaults to current directory.  Set _CONFIG_FOLDER env var to change config location.')
    deploy_parser.set_defaults(func=deploy_vac)



import asyncio
import logging
from typing import Callable, List, Any, Coroutine, Tuple

# Setup basic logging configuration
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class AsyncTaskRunner:
    """
    Example async functions for testing
    ```python
    async def api_call_1(url, params):
        await asyncio.sleep(1)
        if "fail" in params:
            raise ValueError(f"Error in api_call_1 with params: {params}")
        return f"api_call_1 response from {url} with params {params}"

    async def api_call_2(url, params):
        await asyncio.sleep(2)
        if "fail" in params:
            raise ValueError(f"Error in api_call_2 with params: {params}")
        return f"api_call_2 response from {url} with params {params}"

    # Example usage in an existing async function
    async def example_usage():
        runner = AsyncTaskRunner()

        runner.add_task(api_call_1, "http://example.com", {"key": "value"})
        runner.add_task(api_call_2, "http://example.org", {"key": "fail"})

        # Run all tasks within the existing event loop
        results = await runner.run_async()
        for result in results:
            print(result)

    # Example of calling run_sync() in a synchronous context
    if __name__ == "__main__":
        runner = AsyncTaskRunner()

        runner.add_task(api_call_1, "http://example.com", {"key": "value"})
        runner.add_task(api_call_2, "http://example.org", {"key": "fail"})

        # Running in a synchronous context
        results = runner.run_sync()
        for result in results:
            print(result)
    ```

    Example streaming fast and slow
    
    ```python
    import asyncio
    from typing import AsyncGenerator

    # Example streaming function that simulates yielding chunks of data
    async def stream_chunks(url: str, params: dict) -> AsyncGenerator[str, None]:
        # Simulate streaming with a series of chunks
        for i in range(5):
            await asyncio.sleep(1)  # Simulate delay between chunks
            yield f"Chunk {i+1} from {url} with params {params}"

    # Example slow API call function
    async def slow_api_call(url: str, params: dict) -> str:
        await asyncio.sleep(5)  # Simulate a slow API response
        return f"Slow API response from {url} with params {params}"

    # Function to manage streaming and slow API call
    async def process_api_calls(stream_url: str, stream_params: dict, slow_url: str, slow_params: dict) -> AsyncGenerator[str, None]:
        # Create the AsyncTaskRunner instance
        runner = AsyncTaskRunner()

        # Add the slow API call as a task
        runner.add_task(slow_api_call, slow_url, slow_params)

        # Run the slow API call concurrently with the streaming
        slow_api_result_task = asyncio.create_task(runner.run_async())

        # Process the streaming chunks and yield them
        async for chunk in stream_chunks(stream_url, stream_params):
            yield chunk

        # Wait for the slow API call to complete and get the result
        slow_api_results = await slow_api_result_task

        # Yield the slow API response after streaming is finished
        for result in slow_api_results:
            yield result

    # Example usage in an existing async function
    async def example_usage():
        # Define the URLs and parameters for the calls
        stream_url = "http://streaming.example.com"
        stream_params = {"key": "stream_value"}
        slow_url = "http://slowapi.example.com"
        slow_params = {"key": "slow_value"}

        # Process the API calls and stream the results
        async for output in process_api_calls(stream_url, stream_params, slow_url, slow_params):
            print(output)

    # Running the example usage
    if __name__ == "__main__":
        asyncio.run(example_usage())
    ```
    """
    def __init__(self):
        self.tasks = []

    def add_task(self, func: Callable[..., Coroutine], *args: Any):
        """Adds a task (function and its arguments) to the list of tasks to be executed."""
        logger.info(f"Adding task: {func.__name__} with args: {args}")
        self.tasks.append(func(*args))

    def add_group(self, functions: List[Callable[..., Coroutine]], *args: Any):
        """Adds a group of functions that should run in parallel with the same arguments."""
        group_tasks = [func(*args) for func in functions]
        logger.info(f"Adding group of tasks with args: {args}")
        self.tasks.append(asyncio.gather(*group_tasks, return_exceptions=True))

    async def run_async(self) -> List[Any]:
        """Runs all tasks using the current event loop and returns the results."""
        logger.info("Running tasks asynchronously")
        results = await asyncio.gather(*self.tasks, return_exceptions=True)
        self._log_results(results)
        return results

    def run_sync(self) -> List[Any]:
        """Runs all tasks synchronously (blocking) using a new event loop and returns the results."""
        try:
            logger.info("Running tasks synchronously")
            loop = asyncio.get_running_loop()
            return loop.run_until_complete(self.run_async())
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            results = loop.run_until_complete(self.run_async())
            loop.close()
            return results

    def _log_results(self, results: List[Any]):
        """Logs the results of the task executions."""
        for result in results:
            if isinstance(result, Exception):
                logger.error(f"Task resulted in an error: {result}")
            else:
                logger.info(f"Task completed successfully: {result}")




#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from sunholo.custom_logging import log

def document_to_dict(document):
    return {
        "page_content": document.page_content,
        "metadata": document.metadata
    }

def parse_output(bot_output):
    """
    Parses VAC output assuming it has an 'answer' and an optional 'source_documents' key
    
    """
    if isinstance(bot_output, str):

        return {"answer": bot_output}
    
    if isinstance(bot_output, dict) and 'source_documents' in bot_output:
        bot_output['source_documents'] = [document_to_dict(doc) for doc in bot_output['source_documents']]
        if not bot_output.get("answer") or bot_output.get("answer") == "":
            bot_output['answer'] = "(No text was returned)"

        return bot_output
    
    elif isinstance(bot_output, dict) and 'metadata' in bot_output and isinstance(bot_output.get('metadata'), dict) and 'source_documents' in bot_output.get('metadata'):
        metadata = bot_output.get('metadata')
        bot_output['source_documents'] = [document_to_dict(doc) for doc in metadata['source_documents']]
        if not bot_output.get("answer") or bot_output.get("answer") == "":
            bot_output['answer'] = "(No text was returned)"

        return bot_output
    
    elif isinstance(bot_output, dict) and 'output' in bot_output and isinstance(bot_output['output'], dict) and 'content' in bot_output['output']:
        the_output = bot_output['output']

        return {
            'answer': the_output.get('content'),
            'metadata': the_output.get('metadata')
        }
    
    elif isinstance(bot_output, dict):
        if not bot_output.get("answer"):
            raise ValueError(f"VAC output was not a string or a dict with the key 'answer' - got: {bot_output}")
        else:

            return bot_output
        
    else:
        log.error(f"Couldn't parse output for:\n {bot_output}")


# needs to be in minimal to check gcp
import os

import google.auth 
from google.auth.transport import requests
from ..utils.gcp import is_running_on_gcp
from ..custom_logging import log

def get_default_email():

    # https://stackoverflow.com/questions/64234214/how-to-generate-a-blob-signed-url-in-google-cloud-run

    gcs_credentials, project_id = refresh_credentials()

    if gcs_credentials is None:
        log.error("Could not refresh the credentials properly.")
        return None

    service_account_email = getattr(gcs_credentials, 'service_account_email', None)
    # If you use a service account credential, you can use the embedded email
    if not service_account_email:
        service_account_email = os.getenv('GCS_MAIL_USER')
        if not service_account_email:
            log.error("Could not create the credentials for signed requests - no credentials.service_account_email or GCS_MAIL_USER with roles/iam.serviceAccountTokenCreator")
            
            return None
    
    log.info(f"Found default email: {service_account_email=} for {project_id=}")
    return service_account_email, gcs_credentials

def get_default_creds():
    gcs_credentials = None
    project_id = None
    gcs_credentials, project_id = google.auth.default()

    return gcs_credentials, project_id

def refresh_credentials():
    """
    Need to refresh to get a valid email/token for signing URLs from a default service account
    """
    if not is_running_on_gcp():
        log.debug("Not running on Google Cloud so no credentials available for GCS.")
        return None, None
    
    gcs_credentials, project_id = get_default_creds()

    if not gcs_credentials.token or gcs_credentials.expired or not gcs_credentials.valid:
        try:
            r = requests.Request()
            gcs_credentials.refresh(r)

            return gcs_credentials, project_id
        
        except Exception as e:
            log.error(f"Failed to refresh gcs credentials: {e}")

            return None, None
    



import os

try:
    from google.cloud import storage 
except ImportError:
    storage = None

from ..custom_logging import log

def download_files_from_gcs(bucket_name: str, source_folder: str, destination_folder: str=None):
    """
    Download all files from a specified folder in a Google Cloud Storage bucket to a local directory.
    
    Parameters:
    - bucket_name: The name of the GCS bucket.
    - source_folder: The folder (prefix) in the GCS bucket to download files from.
    - destination_folder: The local directory to save the downloaded files, or os.getcwd() if None
    """
    try:
        storage_client = storage.Client()
    except Exception as err:
        log.error(f"Error creating storage client: {str(err)}")
        return None

    # Get the bucket
    bucket = storage_client.bucket(bucket_name)

    # List blobs in the specified folder
    blobs = bucket.list_blobs(prefix=source_folder)

    if not destination_folder:
        destination_folder = os.getcwd()

    # Ensure the destination folder exists
    os.makedirs(destination_folder, exist_ok=True)

    for blob in blobs:
        # Skip if the blob is a directory
        if blob.name.endswith('/'):
            continue
        
        # Define the local path
        local_path = os.path.join(destination_folder, os.path.relpath(blob.name, source_folder))
        
        # Ensure the local folder exists
        os.makedirs(os.path.dirname(local_path), exist_ok=True)

        # Download the blob to a local file
        blob.download_to_filename(local_path)
        log.info(f"Downloaded {blob.name} to {local_path}")